@@ -12,16 +12,25 @@
 //
 // This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 // WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
+//
+// $URL$
+// $Id$
+// SPDX-License-Identifier: GPL-3.0+
+//
+// Author(s)     : Ilker O. Yaz
 
 
 #define CGAL_SURFACE_MESH_SEGMENTATION_H
 
+#include <CGAL/license/Surface_mesh_segmentation.h>
+
+
 #include <CGAL/internal/Surface_mesh_segmentation/Expectation_maximization.h>
 #include <CGAL/internal/Surface_mesh_segmentation/Filters.h>
 #include <CGAL/internal/Surface_mesh_segmentation/Alpha_expansion_graph_cut.h>
 #include <CGAL/internal/Surface_mesh_segmentation/SDF_calculation.h>
 
-#include <CGAL/Mesh_3/dihedral_angle_3.h>
+#include <CGAL/Kernel/global_functions_3.h>
 
 #include <CGAL/property_map.h>
 
@@ -82,7 +91,7 @@ public:
     face_iterator facet_it, fend;
     for(boost::tie(facet_it,fend) = faces(mesh);
         facet_it != fend; ++facet_it) {
-      double sdf_value = sdf_values[*facet_it];
+      double sdf_value = get(sdf_values, *facet_it);
       CGAL_assertion(sdf_value == -1 || sdf_value >= 0); // validity check
       if(sdf_value != -1.0) {
         min_sdf = (std::min)(sdf_value, min_sdf);
@@ -95,7 +104,7 @@ public:
       std::size_t nb_valid_neighbors = 0;
       do {
         if(!(face(opposite(*facet_circulator,mesh),mesh) == boost::graph_traits<Polyhedron>::null_face())) {
-            double neighbor_sdf = sdf_values[face(opposite(*facet_circulator,mesh),mesh)];
+            double neighbor_sdf = get(sdf_values, face(opposite(*facet_circulator,mesh),mesh));
           if(neighbor_sdf != -1) {
             total_neighbor_sdf += neighbor_sdf;
             ++nb_valid_neighbors;
@@ -107,7 +116,7 @@ public:
         still_missing_facets.push_back(*facet_it);
       } else {
         sdf_value = total_neighbor_sdf / nb_valid_neighbors;
-        sdf_values[*facet_it] = sdf_value;
+        put(sdf_values, *facet_it, sdf_value);
         // trying to update min_sdf is pointless, since it is interpolated one of the neighbors sdf will be smaller than it
       }
     }
@@ -116,7 +125,7 @@ public:
     for(typename std::vector<face_descriptor>::iterator it =
           still_missing_facets.begin();
         it != still_missing_facets.end(); ++it) {
-      sdf_values[*it] = min_sdf;
+      put(sdf_values, *it, min_sdf);
     }
   }
 
@@ -128,7 +137,7 @@ public:
     face_iterator facet_it, fend;
     for(boost::tie(facet_it,fend) = faces(mesh);
         facet_it != fend; ++facet_it) {
-      double sdf_value = sdf_values[*facet_it];
+      double sdf_value = get(sdf_values, *facet_it);
       max_sdf = (std::max)(sdf_value, max_sdf);
       min_sdf = (std::min)(sdf_value, min_sdf);
     }
@@ -154,7 +163,7 @@ public:
     face_iterator facet_it, fend;
     for(boost::tie(facet_it,fend) = faces(mesh);
         facet_it != fend; ++facet_it) {
-      sdf_values[*facet_it] = (sdf_values[*facet_it] - min_sdf) / max_min_dif;
+      put(sdf_values, *facet_it, (get(sdf_values, *facet_it) - min_sdf) / max_min_dif);
     }
     return std::make_pair(min_sdf, max_sdf);
   }
@@ -246,7 +255,10 @@ public:
   calculate_sdf_values(double cone_angle, std::size_t number_of_rays,
                        SDFPropertyMap sdf_pmap, bool postprocess_req) {
     // calculate sdf values
-    SDF_calculation_class sdf_calculator(mesh, vertex_point_pmap, false, true, traits);
+    SDF_calculation_class sdf_calculator(mesh, vertex_point_pmap,
+                                         false, /* build_kd_ree */
+                                         true, /* use_diagonal --> set to false to use `AABB_tree::first_intersection()` */
+                                         traits);
     sdf_calculator.calculate_sdf_values(faces(mesh).first, faces(mesh).second,
                                         cone_angle, number_of_rays, sdf_pmap);
 
@@ -292,7 +304,7 @@ public:
     for(boost::tie(facet_it,fend) = faces(mesh);
         facet_it != fend;
         ++facet_it, ++label_it) {
-      segment_pmap[*facet_it] = *label_it; // fill with cluster-ids
+      put(segment_pmap, *facet_it, *label_it); // fill with cluster-ids
     }
     if(clusters_to_segments) {
       // assign a segment id for each facet
@@ -323,7 +335,7 @@ private:
     // As far as I check: if, say, dihedral angle is 5, this returns 175,
     // if dihedral angle is -5, this returns -175.
     // Another words this function returns angle between planes.
-    double n_angle = to_double( ::CGAL::Mesh_3::dihedral_angle(a, b, c, d) );
+    double n_angle = to_double( ::CGAL::approximate_dihedral_angle(a, b, c, d) );
     n_angle /= 180.0;
     bool concave = n_angle > 0;
     double angle = 1 + ((concave ? -1 : +1) * n_angle);
@@ -348,7 +360,7 @@ private:
     face_iterator facet_it, fend;
     for(boost::tie(facet_it,fend) = faces(mesh);
         facet_it != fend; ++facet_it) {
-      double log_normalized = log(sdf_values[*facet_it] * CGAL_NORMALIZATION_ALPHA +
+      double log_normalized = log(get(sdf_values, *facet_it) * CGAL_NORMALIZATION_ALPHA +
                                   1) / log(CGAL_NORMALIZATION_ALPHA + 1);
       normalized_sdf_values.push_back(log_normalized);
     }
@@ -455,7 +467,7 @@ private:
     face_iterator facet_it, fend;
     for(boost::tie(facet_it,fend) = faces(mesh);
         facet_it != fend; ++facet_it) {
-      if(segments[*facet_it] <
+      if(get(segments, *facet_it) <
           number_of_clusters) { // not visited by depth_first_traversal
         double average_sdf_value = breadth_first_traversal(*facet_it, segment_id,
                                    sdf_values, segments);
@@ -482,8 +494,8 @@ private:
     //                        . Then place its sorted index to pmap
     for(boost::tie(facet_it,fend) = faces(mesh);
         facet_it != fend; ++facet_it) {
-      std::size_t segment_id = segments[*facet_it] - number_of_clusters;
-      segments[*facet_it] = segment_id_to_sorted_id_map[segment_id];
+      std::size_t segment_id = get(segments, *facet_it) - number_of_clusters;
+      put(segments, *facet_it, segment_id_to_sorted_id_map[segment_id]);
     }
     return segment_id - number_of_clusters;
   }
@@ -504,10 +516,10 @@ private:
     std::queue<face_descriptor> facet_queue;
     facet_queue.push(root);
 
-    std::size_t prev_segment_id = segments[root];
-    segments[root] = segment_id;
+    std::size_t prev_segment_id = get(segments, root);
+    put(segments, root, segment_id);
 
-    double total_sdf_value = sdf_values[root];
+    double total_sdf_value = get(sdf_values, root);
     std::size_t    visited_facet_count = 1;
 
     while(!facet_queue.empty()) {
@@ -519,11 +531,11 @@ private:
           continue;  // no facet to traversal
         }
         face_descriptor neighbor = face(opposite(*facet_circulator,mesh),mesh);
-        if(prev_segment_id == segments[neighbor]) {
-          segments[neighbor] = segment_id;
+        if(prev_segment_id == get(segments, neighbor)) {
+          put(segments, neighbor, segment_id);
           facet_queue.push(neighbor);
 
-          total_sdf_value += sdf_values[neighbor];
+          total_sdf_value += get(sdf_values, neighbor);
           ++visited_facet_count;
         }
       } while( ++facet_circulator != done);
